//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_PALETTE_DEFINITIONS__
#define __EZ_PALETTE_DEFINITIONS__

//############################################################################
#include "ez_bmp.hpp"
#include "ez_color.hpp"

//############################################################################
class ez_palette : public vector<ez_color>
{
public:
    ez_palette()  {}
    //------------------------------------------------------------------------
    ez_palette(const int& size)
                {
                    resize(size);
                }
    //------------------------------------------------------------------------
    ez_palette(const ez_palette& p)
              : in_use(true   ),
                white (p.white),
                black (p.black),
                index (p.index),
                first (p.first),
                last  (p.last ),
                dud   (       ),
                name  (p.name )
                {
                    clear();
                    reserve(p.size());
                    insert(begin(), p.begin(), p.end());
                }
    //------------------------------------------------------------------------
   ~ez_palette()   {    clear();    }
    //------------------------------------------------------------------------
    ez_palette&   operator = (const ez_palette& p)
                    {
                        reserve(p.size());
                        insert(end(), p.begin(), p.end());
                        white = p.white;
                        black = p.black;
                        index = p.index;
                        first = p.first;
                        last  = p.last ;
                        name  = p.name ;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_palette&   operator += (const ez_palette& p)
                    {
                        reserve(size() + p.size());
                        insert(end(), p.begin(), p.end());
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_color& operator [] (const int& index)
                {
                    if(index >= 0 && index < number_of_colors())
                        return at(index);
                    return dud;
                }
    //------------------------------------------------------------------------
    int    number_of_colors () const {   return (int)size();   }
    //------------------------------------------------------------------------
    void   set_palette_item ( const int&    color_index,
                              const u_char& r,
                              const u_char& g,
                              const u_char& b
                            )
            {
                if(color_index >= 0 && color_index < number_of_colors())
                {
                    at(color_index).r = r;
                    at(color_index).g = g;
                    at(color_index).b = b;
                }
            }
    //------------------------------------------------------------------------
    void   wipe()
            {
                for(int i = 0; i < number_of_colors(); i++)
                    at(i) = (u_char)0x00;
                return;
            }
    //------------------------------------------------------------------------
    ez_palette&   reorder();
    //------------------------------------------------------------------------
    int    from_ifstream_ilda (ifstream& in,  const ez_uail_header& header)       ;
    void   to_ofstream_ilda   (ofstream& out,       ez_uail_header& header) const ;
    //------------------------------------------------------------------------
    int    from_ascii_file       (string file)                     ;
    void   save_as_ascii         (string file)               const ;
    //------------------------------------------------------------------------
    int    from_ezfb_bmp_palette (                         )       ;
    int    to_ezfb_bmp_palette   (                         ) const ;
    int    to_ezfb_bmp_palette   (struct ezfb_bitmap* bmp  ) const ;
    void   find_factors          (                         )       ;
    int    best_match            (ez_color rgb             ) const ;
    void   best_reduction        (                         )       ;
    void   shade                 (u_char shade             )       ;
    void   tint                  (u_char tint              )       ;
    bool   unite                 (const ez_palette& palette)       ;
    //------------------------------------------------------------------------
    bool        in_use;
    
    u_int       white ,
                black ;

    int         index ,
                first ,
                last  ;

    ez_color    dud   ;
    
    string      name  ;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
