//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_uail_gui.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_uail_app::ez_uail_app()
            : ez_uail                 (          )
            , black_threshold         (0         )
            , sample_rate             (48000     ) // Hz
            , lit_dwell_overhang      (2         ) // samples
            , dwell_on_dot            (10        ) // samples
            , output_bmp_size         (1000      ) // pixels square
            , signal_if_not_z         (EZ_SIGNAL_MONO_AVG_ANALOG)
            , unframed_wave_stride    (4800      ) // samples 1/10 of sec
            , lit_distance_delta_max  (600       ) // points
            , blank_distance_delta_max(900       ) // points
            , longest_dwell_microsec  (333       )
            , insignificant_distance  (60        ) // points
            , zero_angle_tolerance    (one_degree) // radians
            , recolor_span_factor     (2.0       )
            , rendered_arc_angle      (6.0       ) // degrees
            , frames_per_second       (16        ) // 0 == every frame is shown exactly once
            , still_frame_duration    (10.0      ) // seconds
            , auto_scale_dxf          (true      )
            , maintain_dxf_origin     (false     )
            , black_dwell_points      (true      )
            , auto_flatten_z          (false     )
            , no_equivalent_vectors   (false     )
            , no_dots                 (false     )
            , invert_wave_output      (false     )
            , auto_optimize           (true      )
            , no_black_match          (false     )
            , allow_lit_black         (false     )
            , dumb_file_format        (false     )
{
    wave_offsets   [0] = 0;
    wave_offsets   [1] = 0;
    wave_offsets   [2] = 0;
    wave_offsets   [3] = 0;
    wave_offsets   [4] = 0;
    wave_offsets   [5] = 0;
    wave_offsets   [6] = 0;
    wave_offsets   [7] = 0;

    signal_bit_mask[0] = 0;
    signal_bit_mask[1] = 0;
    signal_bit_mask[2] = 7;
    signal_bit_mask[3] = 7;
    signal_bit_mask[4] = 7;
    signal_bit_mask[5] = 0;
    signal_bit_mask[6] = 0;
    signal_bit_mask[7] = 0;
}

//############################################################################
void ez_uail_app::apply_wave_offsets(string file_name)
{
    int                      i                     ,
                             span                  ,
                             last                  ,
                             offset_in_max         ,
                             offset_in_min         ,
                             offset_applied_max    ,
                             offset_applied_min    ,
                             offset_applied     [8];
    bool                     inverted       = false;
    ifstream                 wave_ifstream         ;
    ofstream                 wave_ofstream         ;
    ez_wave_sample_pointer   roll_over             ;
    ez_wave_sample_pointer  *sample_window         ;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to rename      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    if(wave_in_header.version == "!LaserBoy!")
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels < 6)
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s is not a LaserBoy wave"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_sample wave_sample(wave_in_header.num_channels);
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_out_header(wave_in_header);
    wave_out_header.ez_wave_mode |= EZ_WAVE_OFFSETS;
    //------------------------------------------------------------------------
    for(i = 0; i < wave_in_header.num_channels; i++)
        offset_applied[i] = wave_in_header.offset[i] - wave_offsets[i];
    //------------------------------------------------------------------------
    offset_in_min      = wave_offsets  [0];
    offset_in_max      = wave_offsets  [0];
    offset_applied_min = offset_applied[0];
    offset_applied_max = offset_applied[0];
    //------------------------------------------------------------------------
    for(i = 0; i < wave_in_header.num_channels; i++)
    {
        if(wave_offsets  [i] > offset_in_max     ) offset_in_max      = wave_offsets  [i];
        if(wave_offsets  [i] < offset_in_min     ) offset_in_min      = wave_offsets  [i];
        if(offset_applied[i] > offset_applied_max) offset_applied_max = offset_applied[i];
        if(offset_applied[i] < offset_applied_min) offset_applied_min = offset_applied[i];
    }    
    //------------------------------------------------------------------------
    span = (offset_applied_max - offset_applied_min) + 1;
    last = span - 1;
    //------------------------------------------------------------------------
    for(i = 0; i < wave_in_header.num_channels; i++)
    {
        wave_out_header.offset[i] = -(offset_in_max - wave_offsets[i]);
        offset_applied[i] = offset_applied_max - offset_applied[i];
    }
    //------------------------------------------------------------------------
    sample_window = new ez_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new ez_wave_sample(wave_in_header.num_channels); // span element array of (6 or 8 channel ez_wave_samples)
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(true)
    {
        if(last) // span > 1
        {
            roll_over = sample_window[0];
            for(i = 0; i < last; i++)
                sample_window[i] = sample_window[i + 1];
            sample_window[last] = roll_over;
        }

        if(!sample_window[last]->from_ifstream_wave(wave_ifstream))
            break;
            
        for(i = 0; i < wave_in_header.num_channels; i++) // each channel
            wave_sample.channel[i] = sample_window[offset_applied[i]]->channel[i];
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header); // write it to disk
    }
    delete sample_window;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    return;    
}

//############################################################################
void ez_uail_app::overwirte_wave_offsets(string file_name)
{
    int                  i                ;
    ifstream             wave_ifstream    ;
    ofstream             wave_ofstream    ;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to rename      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    if(wave_in_header.version == "!LaserBoy!")
    {
        wave_ifstream.close();                                        
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_sample wave_sample(wave_in_header.num_channels);
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_out_header(wave_in_header);
    //------------------------------------------------------------------------
    for(i = 0; i < wave_in_header.num_channels; i++)
        wave_out_header.offset[i] = wave_offsets[i];
    //------------------------------------------------------------------------
    wave_out_header.ez_wave_mode |= EZ_WAVE_OFFSETS;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave(wave_ifstream))
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;    
}

//############################################################################
void ez_uail_app::invert_wave(string file_name)
{
    ifstream  wave_ifstream;
    ofstream  wave_ofstream;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to rename      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    if(wave_in_header.version == "!LaserBoy!")
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_out_header(wave_in_header);
    ez_wave_sample wave_sample    (wave_in_header.num_channels);
    //------------------------------------------------------------------------
    if(wave_in_header.ez_wave_mode & EZ_WAVE_POSITIVE)
        wave_out_header.ez_wave_mode &= ~EZ_WAVE_POSITIVE;
    else
        wave_out_header.ez_wave_mode |= EZ_WAVE_POSITIVE;
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave_inverted(wave_ifstream))
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::invert_signals(string file_name, int polarity[8])
{
    int       i            ;
    ifstream  wave_ifstream;
    ofstream  wave_ofstream;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to rename      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    if(wave_in_header.version == "!LaserBoy!")
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_in_header.ez_wave_mode |= EZ_WAVE_SIGNAL_MATRIX;
    ez_wave_header wave_out_header(wave_in_header);
    ez_wave_sample wave_sample    (wave_in_header.num_channels);
    //------------------------------------------------------------------------
    for(i = 0; i < wave_out_header.num_channels; i++)
        if(polarity[i] == 1)
            wave_out_header.signal_id[i] = -wave_out_header.signal_id[i];
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // take up the space!
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave_polarity(wave_ifstream, polarity))
        wave_sample.to_ofstream_wave(wave_ofstream, wave_out_header);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream);
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::clear_wave_polarity_list(string file_name)
{
    int       i            ;
    ifstream  wave_ifstream;
    ofstream  wave_ofstream;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());
    //------------------------------------------------------------------------
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    if(rename(   (EZ_UAIL_WAVE_SHARE + file_name).c_str()
               , (EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str()
             )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to rename      ").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    //------------------------------------------------------------------------
    if(wave_in_header.version == "!LaserBoy!")
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_ofstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::out | ios::binary);
    if(!wave_ofstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        undo_wave_temp(wave_ifstream, wave_ofstream, file_name);
        return;
    }
    //------------------------------------------------------------------------
    wave_in_header.ez_wave_mode |= EZ_WAVE_SIGNAL_MATRIX;
    ez_wave_header wave_out_header(wave_in_header);
    //------------------------------------------------------------------------
    for(i = 0; i < wave_out_header.num_channels; i++)
        wave_out_header.signal_id[i] = abs(wave_out_header.signal_id[i]);
    //------------------------------------------------------------------------
    wave_out_header.to_ofstream_wave(wave_ofstream); // that's it!
    //------------------------------------------------------------------------
    wave_ifstream.close();
    wave_ofstream.close();
    //------------------------------------------------------------------------
    remove((EZ_UAIL_WAVE_SHARE + "__TEMP__.wav").c_str());    
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::split_wave_XYrgbiLR(string file_name)
{
    int             i               ;
    ifstream        wave_ifstream   ;
    ofstream        wave_ofstream_XY,
                    wave_ofstream_r ,
                    wave_ofstream_g ,
                    wave_ofstream_b ,
                    wave_ofstream_i ,
                    wave_ofstream_LR;
                     
    ez_wave_header  wave_out_header_XY(sample_rate, EZ_WAVE_NO_MODE, 2),
                    wave_out_header_r (sample_rate, EZ_WAVE_NO_MODE, 1),
                    wave_out_header_g (sample_rate, EZ_WAVE_NO_MODE, 1),
                    wave_out_header_b (sample_rate, EZ_WAVE_NO_MODE, 1),
                    wave_out_header_i (sample_rate, EZ_WAVE_NO_MODE, 1),
                    wave_out_header_LR(sample_rate, EZ_WAVE_NO_MODE, 2);

    ez_wave_sample  stereo_sample(2),
                    mono_sample  (1);
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    u_int bmp_bg_color = BMP.make_rgb(AIL->bg_color.r, AIL->bg_color.g, AIL->bg_color.b);
    //------------------------------------------------------------------------    
    if(    file_name.size() < 4
        || file_name.substr(file_name.size() - 4) != ".wav"
      )
        file_name += ".wav";
    //------------------------------------------------------------------------
    wave_ifstream.open((EZ_UAIL_WAVE_SHARE + file_name).c_str(), ios::in | ios::binary);
    if(!wave_ifstream.is_open())
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s failed to open          "
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_header wave_in_header(wave_ifstream);
    if(wave_in_header.version == "!LaserBoy!")
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] error reading %s wave header"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels < 6)
    {
        wave_ifstream.close();                                                   
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s is not a LaserBoy wave"
                    , file_name.c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    ez_wave_sample wave_sample(wave_in_header.num_channels);
    //------------------------------------------------------------------------
    file_name = file_name.substr(0, file_name.size() - 4);
    //------------------------------------------------------------------------
    wave_ofstream_XY.open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_XY.wav").c_str(), ios::out | ios::binary);
    wave_ofstream_r .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_r.wav" ).c_str(), ios::out | ios::binary);
    wave_ofstream_g .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_g.wav" ).c_str(), ios::out | ios::binary);
    wave_ofstream_b .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_b.wav" ).c_str(), ios::out | ios::binary);    
    wave_ofstream_i .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_i.wav" ).c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels == 8)
        wave_ofstream_LR.open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + "_LR.wav").c_str(), ios::out | ios::binary);
    //------------------------------------------------------------------------
    if(    !wave_ofstream_XY.is_open()
        || !wave_ofstream_r .is_open()
        || !wave_ofstream_g .is_open()
        || !wave_ofstream_b .is_open()
        || !wave_ofstream_i .is_open()
        || (    wave_in_header.num_channels == 8
             && !wave_ofstream_LR.is_open()
           )
      )
    {
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , bmp_bg_color
                    , 0
                    , AIL->font_size_factor
                    , "[any key] %s"
                    , (EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name + " failed to open for output").c_str()
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        AIL->process_exception = true;
        return;
    }
    //------------------------------------------------------------------------
    while(wave_sample.from_ifstream_wave(wave_ifstream))
    {
        stereo_sample.channel[0] = wave_sample.channel[0];
        stereo_sample.channel[1] = wave_sample.channel[1];        
        stereo_sample.to_ofstream_wave(wave_ofstream_XY, wave_out_header_XY);
        
        mono_sample.channel[0]   = wave_sample.channel[2];
        mono_sample.to_ofstream_wave(wave_ofstream_r, wave_out_header_r);
        
        mono_sample.channel[0]   = wave_sample.channel[3];
        mono_sample.to_ofstream_wave(wave_ofstream_g, wave_out_header_g);
        
        mono_sample.channel[0]   = wave_sample.channel[4];
        mono_sample.to_ofstream_wave(wave_ofstream_b, wave_out_header_b);
        
        mono_sample.channel[0]   = wave_sample.channel[5];
        mono_sample.to_ofstream_wave(wave_ofstream_i, wave_out_header_i);
        //--------------------------------------------------------------------
        if(wave_in_header.num_channels == 8)
        {
            stereo_sample.channel[0] = wave_sample.channel[6];
            stereo_sample.channel[1] = wave_sample.channel[7];        
            stereo_sample.to_ofstream_wave(wave_ofstream_LR, wave_out_header_LR);
        }
    }
    //------------------------------------------------------------------------
    wave_out_header_XY.to_ofstream_wave(wave_ofstream_XY);
    wave_out_header_r .to_ofstream_wave(wave_ofstream_r );
    wave_out_header_g .to_ofstream_wave(wave_ofstream_g );
    wave_out_header_b .to_ofstream_wave(wave_ofstream_b );
    wave_out_header_i .to_ofstream_wave(wave_ofstream_i );
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels == 8)
        wave_out_header_LR.to_ofstream_wave(wave_ofstream_LR);
    //------------------------------------------------------------------------
    wave_ifstream   .close();
    wave_ofstream_XY.close();
    wave_ofstream_r .close();
    wave_ofstream_g .close();
    wave_ofstream_b .close();
    wave_ofstream_i .close();
    //------------------------------------------------------------------------
    if(wave_in_header.num_channels == 8)
        wave_ofstream_LR.close();    
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::join_XYrgbi_waves(string file_name_XY,
                                    string file_name_r,
                                    string file_name_g,
                                    string file_name_b,
                                    string file_name_i,
                                    string file_name_out
                                   )
{
    ifstream       in_XY,
                   in_r,
                   in_g,
                   in_b,
                   in_i;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    if(    file_name_XY.size() < 4
        || file_name_XY.substr(file_name_XY.size() - 4) != ".wav"
      )
        file_name_XY += ".wav";        
    if(    file_name_r.size() < 4
        || file_name_r.substr(file_name_r.size() - 4) != ".wav"
      )
        file_name_r += ".wav";        
    if(    file_name_g.size() < 4
        || file_name_g.substr(file_name_g.size() - 4) != ".wav"
      )
        file_name_g += ".wav";        
    if(    file_name_b.size() < 4
        || file_name_b.substr(file_name_b.size() - 4) != ".wav"
      )
        file_name_b += ".wav";        
    if(    file_name_i.size() < 4
        || file_name_i.substr(file_name_i.size() - 4) != ".wav"
      )
        file_name_i += ".wav";
    if(    file_name_out.size() < 4
        || file_name_out.substr(file_name_out.size() - 4) != ".wav"
      )
        file_name_out += ".wav";
    //------------------------------------------------------------------------
    in_XY.open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_XY).c_str(), ios::in | ios::binary);
    in_r .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_r ).c_str(), ios::in | ios::binary);
    in_g .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_g ).c_str(), ios::in | ios::binary);
    in_b .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_b ).c_str(), ios::in | ios::binary);
    in_i .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_i ).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(    in_XY.is_open()
        && in_r .is_open()
        && in_g .is_open()
        && in_b .is_open()
        && in_i .is_open()
      )
    {
        ofstream       out;        
        ez_wave_sample sample_XYrgbi(6),
                       sample_XY    (2),
                       sample_r     (1),
                       sample_g     (1),
                       sample_b     (1),
                       sample_i     (1);
        //--------------------------------------------------------------------        
        out.open((EZ_UAIL_WAVE_SHARE + file_name_out).c_str(), ios::out | ios::binary);
        //--------------------------------------------------------------------        
        ez_wave_header wave_in_header_XY(in_XY),
                       wave_in_header_r (in_r ),
                       wave_in_header_g (in_g ),
                       wave_in_header_b (in_b ),
                       wave_in_header_i (in_i ),
                       wave_out_header  (sample_rate,
                                           EZ_WAVE_POSITIVE
                                         | EZ_WAVE_OFFSETS
                                         | EZ_WAVE_SIGNAL_MATRIX
                                         | EZ_SIGNAL_BIT_RESOLUTION,
                                         6
                                        );
        //--------------------------------------------------------------------        
        wave_out_header.offset    [0] = 0;
        wave_out_header.offset    [1] = 0;
        wave_out_header.offset    [2] = 0;
        wave_out_header.offset    [3] = 0;
        wave_out_header.offset    [4] = 0;
        wave_out_header.offset    [5] = 0;

        wave_out_header.signal_id [0] = EZ_SIGNAL_X_POSITION  ;
        wave_out_header.signal_id [1] = EZ_SIGNAL_Y_POSITION  ;
        wave_out_header.signal_id [2] = EZ_SIGNAL_RED_ANALOG  ;
        wave_out_header.signal_id [3] = EZ_SIGNAL_GREEN_ANALOG;
        wave_out_header.signal_id [4] = EZ_SIGNAL_BLUE_ANALOG ;
        wave_out_header.signal_id [5] = EZ_SIGNAL_UNDEFINED   ;
        
        wave_out_header.LSB_tag   [0] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [1] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [2] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [3] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [4] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [5] = EZ_LSB_NOT_USED;

        wave_out_header.resolution[0] = 16;
        wave_out_header.resolution[1] = 16;
        wave_out_header.resolution[2] =  8;
        wave_out_header.resolution[3] =  8;
        wave_out_header.resolution[4] = 16;
        wave_out_header.resolution[5] = 16;
        //--------------------------------------------------------------------
        wave_out_header.to_ofstream_wave(out); // take up the space!
        //--------------------------------------------------------------------
        while(    sample_XY.from_ifstream_wave(in_XY)
               && sample_r .from_ifstream_wave(in_r )
               && sample_g .from_ifstream_wave(in_g )
               && sample_b .from_ifstream_wave(in_b )
               && sample_i .from_ifstream_wave(in_i )
             )
        {
            sample_XYrgbi =   sample_XY
                            + sample_r 
                            + sample_g 
                            + sample_b 
                            + sample_i; // isn't that neat!
                            
            sample_XYrgbi.to_ofstream_wave(out, wave_out_header);
        }
        //--------------------------------------------------------------------
        wave_out_header.to_ofstream_wave(out);
        //--------------------------------------------------------------------
        in_XY.close();
        in_r .close();
        in_g .close();
        in_b .close();
        in_i .close();
        out  .close();
    }
    //------------------------------------------------------------------------
    else
    {
        AIL->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(AIL->bg_color.r,
                                   AIL->bg_color.g,
                                   AIL->bg_color.b
                                  )
                    , 0
                    , AIL->font_size_factor
                    , "[any key] at least one of those files failed to open"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_UAIL_FILE_OPEN_FAILED;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::join_XYikLR_waves(string file_name_XY,
                                    string file_name_i,
                                    string file_name_LR,
                                    string file_name_out
                                   )
{
    ifstream in_XY,
             in_i ,
             in_LR;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    if(    file_name_XY.size() < 4
        || file_name_XY.substr(file_name_XY.size() - 4) != ".wav"
      )
        file_name_XY += ".wav";
    //------------------------------------------------------------------------
    if(    file_name_i.size() < 4
        || file_name_i.substr(file_name_i.size() - 4) != ".wav"
      )
        file_name_i += ".wav";
    //------------------------------------------------------------------------
    if(    file_name_LR.size() < 4
        || file_name_LR.substr(file_name_LR.size() - 4) != ".wav"
      )
        file_name_LR += ".wav";        
    //------------------------------------------------------------------------
    if(    file_name_out.size() < 4
        || file_name_out.substr(file_name_out.size() - 4) != ".wav"
      )
        file_name_out += ".wav";
    //------------------------------------------------------------------------
    in_XY.open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_XY).c_str(), ios::in | ios::binary);
    in_i .open((EZ_UAIL_UNFORMATTED_WAVE_SHARE + file_name_i ).c_str(), ios::in | ios::binary);
    in_LR.open((EZ_UAIL_AUDIO_SHARE      + file_name_LR).c_str(), ios::in | ios::binary);
    //------------------------------------------------------------------------
    if(    in_XY.is_open()
        && in_i .is_open()
        && in_LR.is_open()
      )
    {
        ofstream       out;        
        ez_wave_sample sample_XYikLR(6),
                       sample_XY    (2),
                       sample_i     (1),
                       sample_k     (1),
                       sample_LR    (2);
        //--------------------------------------------------------------------        
        out.open((EZ_UAIL_WAVE_SHARE + file_name_out).c_str(), ios::out | ios::binary);
        //--------------------------------------------------------------------        
        ez_wave_header wave_in_header_XY(in_XY),
                       wave_in_header_i (in_i ),
                       wave_in_header_LR(in_LR),
                       wave_out_header  (sample_rate,
                                           EZ_WAVE_POSITIVE
                                         | EZ_WAVE_OFFSETS
                                         | EZ_WAVE_SIGNAL_MATRIX
                                         | EZ_SIGNAL_BIT_RESOLUTION,
                                         6
                                        );
        //--------------------------------------------------------------------        
        wave_out_header.offset    [0] = 0;
        wave_out_header.offset    [1] = 0;
        wave_out_header.offset    [2] = 0;
        wave_out_header.offset    [3] = 0;
        wave_out_header.offset    [4] = 0;
        wave_out_header.offset    [5] = 0;

        wave_out_header.signal_id [0] = EZ_SIGNAL_X_POSITION ;
        wave_out_header.signal_id [1] = EZ_SIGNAL_Y_POSITION ;
        wave_out_header.signal_id [2] = EZ_SIGNAL_MONO_ANALOG;
        wave_out_header.signal_id [3] = EZ_SIGNAL_MONO_TTL   ;
        wave_out_header.signal_id [4] = EZ_SIGNAL_AUDIO_LEFT ;
        wave_out_header.signal_id [5] = EZ_SIGNAL_AUDIO_RIGHT;
        
        wave_out_header.LSB_tag   [0] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [1] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [2] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [3] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [4] = EZ_LSB_NOT_USED;
        wave_out_header.LSB_tag   [5] = EZ_LSB_NOT_USED;
        
        wave_out_header.resolution[0] = 16;
        wave_out_header.resolution[1] = 16;
        wave_out_header.resolution[2] =  8;
        wave_out_header.resolution[3] =  8;
        wave_out_header.resolution[4] = 16;
        wave_out_header.resolution[5] = 16;
        //--------------------------------------------------------------------
        wave_out_header.to_ofstream_wave(out); // take up the space!
        //--------------------------------------------------------------------
        while(    sample_XY.from_ifstream_wave(in_XY)
               && sample_i .from_ifstream_wave(in_i )
               && sample_LR.from_ifstream_wave(in_LR)
             )
        {
            sample_k.channel[0] = (   (sample_i.channel[0] > (EZ_MAX_COLOR_SHORT / 4)) 
                                    ? (EZ_MAX_COLOR_SHORT) 
                                    : (0)
                                  );

            sample_XYikLR =   sample_XY
                            + sample_i
                            + sample_k                            
                            + sample_LR;
                            
            sample_XYikLR.to_ofstream_wave(out, wave_out_header);
        }
        //--------------------------------------------------------------------
        wave_out_header.to_ofstream_wave(out);
        //--------------------------------------------------------------------
        in_XY.close();
        in_i .close();
        in_LR.close();
        out  .close();
    }
    //------------------------------------------------------------------------
    else
    {
        AIL->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(AIL->bg_color.r,
                                   AIL->bg_color.g,
                                   AIL->bg_color.b
                                  )
                    , 0
                    , AIL->font_size_factor
                    , "[any key] at least one of those files failed to open"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_UAIL_FILE_OPEN_FAILED;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::add_audio_to_wave(string file_name_laser,
                                    string file_name_audio,
                                    string file_name_out
                                   )
{
    ifstream  in_laser,
              in_audio;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    if(    file_name_laser.size() < 4
        || file_name_laser.substr(file_name_laser.size() - 4) != ".wav"
      )
        file_name_laser += ".wav";   
    //------------------------------------------------------------------------
    if(    file_name_out.size() < 4
        || file_name_out.substr(file_name_out.size() - 4) != ".wav"
      )
        file_name_out += ".wav";
    //------------------------------------------------------------------------
    if(file_name_audio == "_0")
    {
        in_laser.open((EZ_UAIL_WAVE_SHARE  + file_name_laser).c_str(), ios::in | ios::binary);
        if(in_laser.is_open())
        {
            ez_wave_header wave_in_header_laser(in_laser);    
            if(wave_in_header_laser.version != "!LaserBoy!")
            {
                ofstream       out;        
                                       
                ez_wave_sample sample_laser      (6), // 6 chennels
                               sample_audio      (2), // 2 chennels (zero)
                               sample_laser_audio(8); // 8 chennels
                //------------------------------------------------------------
//                sample_audio.channel[1] = EZ_MAX_SHORT; // right channel
                //------------------------------------------------------------
                ez_wave_header wave_out_header(wave_in_header_laser, 0); // 8ch.
                //------------------------------------------------------------
                out.open((EZ_UAIL_WAVE_SHARE + file_name_out).c_str(), ios::out | ios::binary);
                wave_out_header.to_ofstream_wave(out); // take up the space!
                //------------------------------------------------------------
                while(sample_laser.from_ifstream_wave(in_laser))
                {
                    sample_laser_audio = sample_laser + sample_audio; // isn't that neat!
                    sample_laser_audio.to_ofstream_wave(out, wave_out_header);
                }
                //------------------------------------------------------------
                wave_out_header.to_ofstream_wave(out);
                //------------------------------------------------------------
                out.close();
            }
            else
            {
                AIL->process_exception = true;
                bmp_printf(   &BMP
                            , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                            , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                            , BMP.white
                            , BMP.make_rgb(AIL->bg_color.r,
                                           AIL->bg_color.g,
                                           AIL->bg_color.b
                                          )
                            , 0
                            , AIL->font_size_factor
                            , "[any key] %s is not a LaserBoy wave file."
                            , file_name_laser.c_str()
                          );
                bmp_to_SDL_Surface(&BMP, screen);
                error = EZ_UAIL_FILE_OPEN_FAILED;
            }
            in_laser.close();
        }
        //--------------------------------------------------------------------
        else
        {
            AIL->process_exception = true;
            bmp_printf(   &BMP
                        , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                        , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                        , BMP.white
                        , BMP.make_rgb(AIL->bg_color.r,
                                       AIL->bg_color.g,
                                       AIL->bg_color.b
                                      )
                        , 0
                        , AIL->font_size_factor
                        , "[any key] %s failed to open"
                        , file_name_laser.c_str()
                      );
            bmp_to_SDL_Surface(&BMP, screen);
            error = EZ_UAIL_FILE_OPEN_FAILED;
        }
        //--------------------------------------------------------------------
    }
    else
    {   
        if(    file_name_audio.size() < 4
            || file_name_audio.substr(file_name_audio.size() - 4) != ".wav"
          )
            file_name_audio += ".wav";        
        //------------------------------------------------------------------------
        in_laser.open((EZ_UAIL_WAVE_SHARE  + file_name_laser).c_str(), ios::in | ios::binary);
        in_audio.open((EZ_UAIL_AUDIO_SHARE + file_name_audio).c_str(), ios::in | ios::binary);
        if(    in_laser.is_open()
            && in_audio.is_open()
          )
        {
            ez_wave_header wave_in_header_laser(in_laser),
                           wave_in_header_audio(in_audio);
    
            if(wave_in_header_laser.version != "!LaserBoy!")
            {
                ofstream       out;        
                                       
                ez_wave_sample sample_laser      (6), // 6 chennels
                               sample_audio      (2), // 2 chennels
                               sample_laser_audio(8); // 8 chennels
        
                ez_wave_header wave_out_header   (wave_in_header_laser, 0); // 8ch.
                //----------------------------------------------------------------
                out.open((EZ_UAIL_WAVE_SHARE + file_name_out).c_str(), ios::out | ios::binary);
                wave_out_header.to_ofstream_wave(out); // take up the space!
                //----------------------------------------------------------------
                while(    sample_laser.from_ifstream_wave(in_laser)
                       && sample_audio.from_ifstream_wave(in_audio)
                     )
                {
                    sample_laser_audio = sample_laser + sample_audio; // isn't that neat!
                    sample_laser_audio.to_ofstream_wave(out, wave_out_header);
                }
                //----------------------------------------------------------------
                wave_out_header.to_ofstream_wave(out);
                //----------------------------------------------------------------
                out.close();
            }
            else
            {
                AIL->process_exception = true;
                bmp_printf(   &BMP
                            , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                            , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                            , BMP.white
                            , BMP.make_rgb(AIL->bg_color.r,
                                           AIL->bg_color.g,
                                           AIL->bg_color.b
                                          )
                            , 0
                            , AIL->font_size_factor
                            , "[any key] %s is not a LaserBoy wave file."
                            , file_name_laser.c_str()
                          );
                bmp_to_SDL_Surface(&BMP, screen);
                error = EZ_UAIL_FILE_OPEN_FAILED;
            }
            in_laser.close();
            in_audio.close();
        }
        //------------------------------------------------------------------------
        else
        {
            AIL->process_exception = true;
            bmp_printf(   &BMP
                        , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                        , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                        , BMP.white
                        , BMP.make_rgb(AIL->bg_color.r,
                                       AIL->bg_color.g,
                                       AIL->bg_color.b
                                      )
                        , 0
                        , AIL->font_size_factor
                        , "[any key] at least one of those files failed to open"
                      );
            bmp_to_SDL_Surface(&BMP, screen);
            error = EZ_UAIL_FILE_OPEN_FAILED;
        }
        //------------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_uail_app::call_frame_effect(string name)
{
    int i;
    //------------------------------------------------------------------------
    AIL->process_exception = false;
    //------------------------------------------------------------------------
    for(i = 0; i < (int)frame_effect_names.size(); i++)
        if(frame_effect_names[i] == name)
            break;
    //------------------------------------------------------------------------            
    if(i < (int)frame_effect_names.size())
        frame_effects[i](&(current_frame()));
    //------------------------------------------------------------------------
    else
    {
        AIL->process_exception = true;
        bmp_printf(   &BMP
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , EZ_STATS_DISPLAY_OFFSET * AIL->font_size_factor
                    , BMP.white
                    , BMP.make_rgb(AIL->bg_color.r,
                                   AIL->bg_color.g,
                                   AIL->bg_color.b
                                  )
                    , 0
                    , AIL->font_size_factor
                    , "[any key] no frame effect of that name found"
                  );
        bmp_to_SDL_Surface(&BMP, screen);
        error = EZ_EFFECT_NAME_NOT_FOUND;
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_uail_app::call_frame_set_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        if(frame_set_effect_names[i] == name)
            break; 
    if(i < (int)frame_set_effect_names.size())
        frame_set_effects[i](this);
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
